/*
  FCrypt - A censorship-resistant cryptosystem
  Copyright (C) 2012  Fabio J. Gonzalez <gonzalfj@ymail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include "arguments.h"
#include "common.h"
#include <gcrypt.h>
#include <unistd.h>
#include <glib.h>
#include <glib/gi18n.h>
#include "config.h"
#include <signal.h>

#define KILO(x)					\
  ((x) * 1024)

#define MEGA(x)					\
  (KILO(x) * 1024)

#define GIGA(x)					\
  (MEGA(x) * 1024)
  
#define TERA(x)					\
  (GIGA(x) * 1024)

int act_running;

/* Size of random seed */
#define random_size 32

/* How many random number use
   with generating a new seed?
   In some cases this is an
   approximate value. */
gsize random_dur = MEGA(5);

/* Structure used to generate
   random numbers */
struct random_data_struct
{
  /*The random number generated*/
  guint8 out[random_size];

  /* How many bytes were used?
     We need a new seed? */
  gsize used;

  /* A random seed readed from:
     /dev/random */
  guint8 seed[random_size];
};

struct random_data_struct random_data;

typedef struct
{
  guint64 key;
  gsize pos;
  gsize len;
  char password[32];
} Sector;

/* HEADER_SECTOR_SIZE is a best name */
#define SECTOR_SIZE sizeof(Sector)

/*In sectors(view SECTOR_SIZE)*/
#define HEADER_SIZE MEGA(1)

#define CSECTOR_SIZE (fgcry_len(SECTOR_SIZE))

gsize
file_len(FILE* fp);

#define str(x) str2(x)
#define str2(x) #x

#define fgcry_err(x)					\
  if(x != 0)						\
    {							\
      printf( "%s\n", gcry_strerror(x) );		\
      printf( "source:%s\n", gcry_strsource(x) );	\
      fflush(stdout);					\
      g_assert(x == 0);					\
    }

#define random_message					\
  printf("\nGenerating random data.\n"			\
	 "For it works, you need do\n"			\
	 "something(eg. move the mouse randomly or\n"	\
	 "or press keys randomly)\n");			\
  fflush(stdout)

#define show(x, ...)				\
  printf(x);					\
  fflush(stdout);

FILE *lrandom_file;

/* When it needs to fill a section with
   random numbers, print completed
   percentage, and waits for "entropy". */

#define rand_comp(i, j) \
  _rand_comp( (i), (j) )

void 
_rand_comp( gulong i, gulong j ) 
{
  if( i % 60000 == 0 )
    {
      printf("%.3f ", 100. / j * i);
      fflush(stdout);
    }
}

gcry_error_t
fgcry_decrypt( gcry_cipher_hd_t* cipher, char* password,
	       size_t password_len, void* data,
	       size_t len );

gcry_error_t
fgcry_encrypt( gcry_cipher_hd_t* cipher, char* password,
	       size_t password_len, void* data,
	       size_t len );

size_t
fgcry_len( size_t len );

#if HAVE_GTHREAD
void
auto_entropy()
{
  guint8 b;
  gsize i;

  while(1)
    {
      for( i = 0; i < random_size; i++)
	{
	  fread( &b, 1, 1, lrandom_file);
	  random_data.seed[i] ^= b;
	}

      random_data.used = 0;

      for( i = 0; i < random_size; i++)
	{
	  fread( &b, 1, 1, lrandom_file);
	  random_data.out[i] ^= b;
	}
    }
}
#endif

void stop()
{
  act_running = 0;
  g_print("Stoping...\n");
}

gsize
file_len(FILE* fp)
{
  gsize pos, ret;

  pos = ftell(fp);
  fseek( fp, 0, SEEK_END );
  ret = ftell(fp);
  fseek( fp, pos, SEEK_SET );

  return ret;
}

/* The following functions read and write to
   the file. Similar to fread and fwrite. But
   improving the output of the randomly seed
   with the data found in the archives. */
gsize
rand_fread( guint8* buf, gsize size,
	    gsize count, FILE* fp )
{
  gsize pos, i, j;
  guint8 byte;
  gsize ret;

  /* Read a random position of the file
     instead of reading this (should be
     slower). */
  if( f_arg_exists("paranoid") &&
      file_len(fp) > random_size )
    {
      pos = ftell(fp);

      j = random_data.out[random_data.seed[3] % random_size]
	% (file_len(fp)-random_size);

      fseek( fp, j, SEEK_SET );

      for( i = 0; i < random_size; i++ )
	{
	  fread( byte, 1, sizeof(byte), fp );
	  random_data.out[i] ^= byte;
	}

      fseek( fp, pos, SEEK_SET ); 
    }

  ret = fread( buf, size, count, fp );

  for( i = 0; i < size; i++ )
      random_data.out[i] ^= buf[i];

  return ret;
}

gsize
rand_fwrite( guint8* buf, gsize size,
	    gsize count, FILE* fp )
{
  gsize pos, i, j;
  guint8 byte;
  gsize ret;

  /* Read a random position of the file
     instead of reading this (should be
     slower). */
  if( f_arg_exists("paranoid") &&
      file_len(fp) > random_size )
    {
      pos = ftell(fp);

      j = random_data.out[random_data.seed[3] % random_size]
	% (file_len(fp)-random_size);

      fseek( fp, j, SEEK_SET );

      for( i = 0; i < random_size; i++ )
	{
	  fread( byte, 1, sizeof(byte), fp );
	  random_data.out[i] ^= byte;
	}

      fseek( fp, pos, SEEK_SET ); 
    }

  ret = fwrite( buf, size, count, fp );

  for( i = 0; i < size; i++ )
      random_data.out[i] ^= buf[i];

  return ret;
}

void
gen_random_seed()
{
  guint8 tmp;
  gsize i, j, k;

  for( i = 0; i < random_size; i++ )
    {
      fread( &tmp, 1, 1, lrandom_file );
      random_data.seed[i] ^= tmp;
    }
}

/* Does not generate new seeds, generates the
   amount of data that can with this seed.
   When the seed is fully depleted, it leaves
   intact the buffer and returns 0. Always
   returns the amount of random numbers that
   could generate. */
gsize
try_getrandom( guint8* data, gsize len )
{
  gsize i, j, k;
  guint8 tmp[random_size];
  gcry_error_t gcry_err;

  /*We must generate a new seed?*/
  if( random_data.used >= random_dur )
    return 0;

  if( f_arg_exists("paranoid") )
    {
      j = random_size / 4;
      if( j == 0 )
	j = 1;
    }
  else
    j = random_size;

  for( i = 0; i < len; i += j )
    {
      gcry_err =
	gcry_kdf_derive( &(random_data.seed[0]),
			 random_size, 
			 GCRY_KDF_SIMPLE_S2K,
			 GCRY_MD_SHA512, NULL, 0,
			 0, random_size, &tmp[0] );
      fgcry_err(gcry_err);

      for( k = 0; k < random_size; k++ )
	random_data.seed[k] ^= tmp[k];

      gcry_err =
	gcry_kdf_derive( &(random_data.seed[0]),
			 random_size, 
			 GCRY_KDF_SIMPLE_S2K,
			 GCRY_MD_SHA512, NULL, 0,
			 0, random_size, &tmp[0] );
      fgcry_err(gcry_err);

      for( k = 0; k < random_size; k++ )
	random_data.out[k] ^= tmp[k];

      for( k = i; k < i+j; k++ )
	data[k] = random_data.out[k];

      random_data.used += j;
      if( random_data.used >= random_dur )
	return i+j;
    }
}

void
getrandom( guint8* data, gsize len )
{
  gsize i, j, k;
  guint8 tmp[random_size];
  gcry_error_t gcry_err;

  /*We must generate a new seed?*/
  if( random_data.used >= random_dur )
    gen_random_seed();

  if( f_arg_exists("paranoid") )
    {
      j = random_size / 3;
      if( j == 0 )
	j = 1;
    }
  else
    j = random_size;

  for( i = 0; i < len; i += j )
    {
      gcry_err =
	gcry_kdf_derive( &(random_data.seed[0]),
			 random_size, 
			 GCRY_KDF_SIMPLE_S2K,
			 GCRY_MD_SHA1, NULL, 0,
			 0, random_size, &tmp[0] );
      fgcry_err(gcry_err);

      for( k = 0; k < random_size; k++ )
	random_data.seed[k] ^= tmp[k];

      gcry_err =
	gcry_kdf_derive( &(random_data.seed[0]),
			 random_size, 
			 GCRY_KDF_SIMPLE_S2K,
			 GCRY_MD_SHA224, NULL, 0,
			 0, random_size, &tmp[0] );
      fgcry_err(gcry_err);

      for( k = 0; k < random_size; k++ )
	random_data.out[k] ^= tmp[k];

      for( k = i; k < i+j; k++ )
	data[k] = random_data.out[k];

      random_data.used += j;
      if( random_data.used >= random_dur )
	gen_random_seed();
    }
}

gsize
fcrypt_read( FILE* fp, Sector* sec, guint8* data,
	     size_t pos, size_t len )
{
  size_t block_size = 
    gcry_cipher_get_algo_blklen(GCRY_CIPHER_AES256);
  guint8* block = g_malloc(block_size);
  size_t i, j, k;
  size_t t_pos, t_len;
  gcry_error_t gcry_err;
  gcry_cipher_hd_t cipher;
  fArgsInfo* info = f_data_get (0, "args_info");
  gsize readed;

  gcry_err =
    gcry_cipher_open( &cipher, GCRY_CIPHER_AES256, 
		      GCRY_CIPHER_MODE_ECB, 0 );
  if(gcry_err != 0)
	goto err;

  gcry_err =
    gcry_cipher_setkey( cipher, &(sec->password[0]), 32);
  if(gcry_err != 0)
    goto err;

  t_pos = (pos / block_size) * block_size;

  /* (pos % block_size) is where I start
     to read, but also are all the bytes
     that I could not read. */
  t_len = len + (pos % block_size);

  /* If not a multiple of the size of each block,
     I need to bring more than a block represents,
     if not lacked space instead of leftovers. */
  if( t_len % block_size != 0 )
    t_len = ((t_len / block_size)+1) * block_size;

  j = t_len / block_size;

  for( ; pos + (j * block_size) > sec->len; j-- );

  t_len = t_len * j;

  /* We'll have to read the appropriate position
     of the file. If not, we get false results
     while decrypting. */
  fseek( fp, sec->pos + t_pos, SEEK_SET );
  
  /*
    As we are in the right place File subtract it from
    the position. We have the position to read the first
    block.
    It is the same as: 
    pos -= (pos / block_size * block_size)
    We do not need subtract the sector begin(sec->pos),
    because pos is relative to the begin of sector(sec->pos)
  */
  pos = pos % block_size;

  for( i = 0; i < j; i++ )
    {
      if( f_arg_exists("better-random") )
	rand_fread( block, 1, block_size, fp );
      else
	fread( block, 1, block_size, fp );

      if( fgcry_decrypt( &cipher, NULL, 0,
			 block, block_size) != 0 )
	{
	  printf( _("error to decrypt data\n"
		    "This is probably a BUG.\n"
		    "I suggest report at <%s>\n"),
		  info->report_bugs );
	  return i * block_size;
	}

      for( k = pos - (i * block_size); k < block_size &&
	     k+(i*block_size) < len; k++ )
	  data[k+(i*block_size)] = block[k];

      readed += k;
      pos += k;
    }

  g_free(block);
  gcry_cipher_close(cipher);
  return readed;

 err:
  g_free(block);
  gcry_cipher_close(cipher);
  return 0;
}

gsize
fcrypt_write( FILE* fp, Sector* sec, guint8* data,
	      size_t pos, size_t len )
{
  size_t block_size = 
    gcry_cipher_get_algo_blklen(GCRY_CIPHER_AES256);
  guint8* block = g_malloc(block_size);
  size_t i, j, k;
  size_t t_pos, t_len;
  gcry_error_t gcry_err;
  gcry_cipher_hd_t cipher;
  fArgsInfo* info = f_data_get (0, "args_info");
  gsize readed;
  gsize *max;

  gcry_err =
    gcry_cipher_open( &cipher, GCRY_CIPHER_AES256, 
		      GCRY_CIPHER_MODE_ECB, 0 );
  if(gcry_err != 0)
	goto err;

  gcry_err =
    gcry_cipher_setkey( cipher, &(sec->password[0]), 32);
  if(gcry_err != 0)
    goto err;

  t_pos = (pos / block_size) * block_size;

  /* (pos % block_size) is where I start
     to read, but also are all the bytes
     that I could not read. */
  t_len = len + (pos % block_size);

  /* If not a multiple of the size of each block,
     I need to bring more than a block represents,
     if not lacked space instead of leftovers. */
  if( t_len % block_size != 0 )
    t_len = ((t_len / block_size)+1) * block_size;

  j = t_len / block_size;

  for( ; pos + (j * block_size) > sec->len; j-- );

  t_len = t_len * j;

  /* We'll have to read the appropriate position
     of the file. If not, we get false results
     while decrypting. */
  fseek( fp, sec->pos + t_pos, SEEK_SET );
  
  /*
    As we are in the right place File subtract it from
    the position. We have the position to read the first
    block.
    It is the same as: 
    pos -= (pos / block_size * block_size)
    We do not need subtract the sector begin(sec->pos),
    because pos is relative to the begin of sector(sec->pos)
  */
  pos = pos % block_size;

  for( i = 0; i < j; i++ )
    {
      if( f_arg_exists("better-random") )
	rand_fread( block, 1, block_size, fp );
      else
	fread( block, 1, block_size, fp );

      if( fgcry_decrypt( &cipher, NULL, 0,
			 block, block_size) != 0 )
	{
	  printf( _("error to decrypt data\n"
		    "This is probably a BUG.\n"
		    "I suggest report at <%s>\n"),
		  info->report_bugs );
	  return i * block_size;
	}

      for( k = pos - (i * block_size); k < block_size &&
	     k+(i*block_size) < len; k++ )
	  block[k] = data[k+(i*block_size)];
	
      pos += k;

      if( fgcry_encrypt( &cipher, NULL, 0,
			 block, block_size ) != 0 )
	{
	  printf( _("error to encrypt data\n"
		    "This is probably a BUG.\n"
		    "I suggest report at <%s>\n"),
		  info->report_bugs );
	  return (i * block_size) +
	    ( k - ( pos - (i*block_size) ) );
	}

      if(f_arg_exists("better-random") )
	rand_fwrite( block, 1, block_size, fp);
      else
	fwrite( block, 1, block_size, fp );

      readed += k;
    }

  g_free(block);
  gcry_cipher_close(cipher);
  return readed;

 err:
  g_free(block);
  gcry_cipher_close(cipher);
  return 0;
}

gcry_error_t
fgcry_encrypt( gcry_cipher_hd_t* cipher, char* password,
	       size_t password_len, void* data,
	       size_t len )
{
  gpointer hpass;
  gcry_error_t gcry_err;
  size_t block_size = 
    gcry_cipher_get_algo_blklen(GCRY_CIPHER_AES256);
  size_t i, old_len;
  gcry_cipher_hd_t _cipher;
  int delete = 0;

  hpass = g_malloc0(32);

  if( cipher == NULL )
    {
      cipher = &_cipher;

      gcry_err =
	gcry_cipher_open( cipher, GCRY_CIPHER_AES256, 
			  GCRY_CIPHER_MODE_ECB, 0);

      if(gcry_err != 0)
	goto end;

      delete = (!0);
    }

  if(password != NULL )
    {
      gcry_err =
	gcry_kdf_derive( password, strlen(password),
			 GCRY_KDF_SIMPLE_S2K,
			 GCRY_MD_SHA256, NULL,
			 0, 0, 32,
			 hpass );

      if(gcry_err != 0)
	goto end;

      gcry_err =
	gcry_cipher_setkey(*cipher, hpass, 32);

      if(gcry_err != 0)
	goto end;
    
    }

  if( len % block_size != 0 )
    {
      g_print("Invalid size!\n");
      return -1;
    }

  for( i = 0; i < len; i += block_size )
    {
      gcry_err =
	gcry_cipher_encrypt( *cipher, data + i,
			     block_size, NULL, 0);

      if( gcry_err != 0 )
	goto end;
    }

 end:
  if(delete)
    gcry_cipher_close(_cipher);
  g_free(hpass);

  return gcry_err;
}

Sector
fgcry_getsector( FILE* fp, char* password,
		 size_t pass_len )
{
  Sector ret;
  unsigned long long key;
  size_t i;
  size_t sec_len = fgcry_len(sizeof(Sector));
  Sector* sec = g_malloc( sec_len );
  gcry_error_t gcry_err;

  fseek(fp, 0, SEEK_SET);

  gcry_err =
    gcry_kdf_derive( password, pass_len, 
		     GCRY_KDF_SIMPLE_S2K,
		     GCRY_MD_MD5, NULL, 0,
		     0, sizeof(key), &key );
  fgcry_err(gcry_err);

  for( i = 0; i < HEADER_SIZE; i++ )
    {
      fread( sec, 1, sec_len, fp );

      fgcry_decrypt( NULL, password, 
		     pass_len, sec, sec_len );

      if( sec->key == key )
	break;
    }

  if( i == HEADER_SIZE )
    {
      ret.key = 0;
      return ret;
    }

  ret = *sec;

  g_free(sec);

  return ret;
}

gboolean
fgcry_delsector( FILE* fp, char* password,
		 size_t pass_len )
{
  unsigned long long key;
  size_t i;
  size_t sec_len = fgcry_len(sizeof(Sector));
  Sector* sec = g_malloc( sec_len );
  gcry_error_t gcry_err;

  fseek(fp, 0, SEEK_SET);

  gcry_err =
    gcry_kdf_derive( password, pass_len, 
		     GCRY_KDF_SIMPLE_S2K,
		     GCRY_MD_MD5, NULL, 0,
		     0, sizeof(key), &key );
  fgcry_err(gcry_err);

  for( i = 0; i < HEADER_SIZE; i++ )
    {
      fread( sec, 1, sec_len, fp );

      gcry_err = fgcry_decrypt( NULL, password, 
				pass_len, sec, sec_len );
      fgcry_err(gcry_err);

      if( sec->key == key )
	break;
    }

  if( i == HEADER_SIZE )
    goto err;

  fseek( fp, i * sec_len, SEEK_SET);
  getrandom( sec, sec_len);
  if( f_arg_exists("better-random") )
    rand_fwrite( sec, 1, sec_len, fp );
  else
    fwrite( sec, 1, sec_len, fp);

  g_free(sec);
  return TRUE;

 err:
  g_free(sec);
  return FALSE;
}

gcry_error_t
fgcry_decrypt( gcry_cipher_hd_t* cipher, char* password,
	       size_t password_len, void* data,
	       size_t len )
{
  gpointer hpass;
  size_t block_size = 
    gcry_cipher_get_algo_blklen(GCRY_CIPHER_AES256);
  size_t i, old_len;
  gcry_cipher_hd_t _cipher;
  int delete = 0;
  unsigned long new_len;
  void* new_data;
  gboolean have_data; /*have_new_data a better name*/
  gcry_error_t gcry_err;

  hpass = g_malloc0(32);

  if( cipher == NULL )
    {
      cipher = &_cipher;

      gcry_err =
	gcry_cipher_open( cipher, GCRY_CIPHER_AES256, 
			  GCRY_CIPHER_MODE_ECB, 0);

      if(gcry_err != 0)
	goto end;

      delete = (!0);
    }

  if(password != NULL )
    {
      gcry_err =
	gcry_kdf_derive( password, strlen(password),
			 GCRY_KDF_SIMPLE_S2K,
			 GCRY_MD_SHA256, NULL,
			 0, 0, 32,
			 hpass );

      if(gcry_err != 0)
	goto end;

      gcry_err =
	gcry_cipher_setkey(*cipher, hpass, 32);

      if(gcry_err != 0)
	goto end;
    
    }

  if( len % block_size != 0 )
    {
      g_print("Invalid size!\n");
      return -1;
    }

  for( i = 0; i < len; i += block_size )
    {
      gcry_err =
	gcry_cipher_decrypt( *cipher, data + i,
			     block_size, NULL, 0);

      if( gcry_err != 0 )
	goto end;
    }

 end:
  if(delete)
    gcry_cipher_close(_cipher);
  g_free(hpass);

  return gcry_err;
}

size_t
fgcry_len( size_t len )
{
  size_t block_size = 
    gcry_cipher_get_algo_blklen(GCRY_CIPHER_AES256);

  if( len % block_size == 0)
    return len;

  return block_size * ((len / block_size)+1);  
}

gpointer
fgcry_newbuffer( gpointer old_buffer, gulong old_len,
		 gboolean free_buffer )
{
  size_t len = fgcry_len(old_len);
  gpointer ret;

  if(len == old_len)
    return old_buffer;

  ret = g_malloc( len );

  g_memmove( ret, old_buffer, old_len );

  /* Often it is best to avoid giving 
     hints about the contents. */
  if( f_arg_exists("paranoid") )
    getrandom( ret + old_len, len - old_len);

  if(free_buffer)
    g_free(old_buffer);

  return ret;
}

/* Remember that all operations and file  encryption are
   done in user-mode by this program. Each request
   for /dev/fcrypt driver, "calls" this program during
   its execution (of both) to send the requested data. */
int 
main( int argc, char* argv[] ) 
{
  char* filename;
  FILE *fp;

  /*Arbitrary data*/
  char *tmp, *c, *c2;

  /*Arbitrary indexes*/
  gsize i, j, k, l;
  guint8 byte;
  Sector* sec = g_malloc0( fgcry_len(sizeof(Sector)) );
  Request* driver_req = malloc(sizeof(Request) + 1025);
  Response* driver_res = malloc(sizeof(Response) + 1025);

  /*Arbitrary encryption*/
  gcry_md_hd_t hash;
  gcry_cipher_hd_t cipher;

  /* Interface that this program communicates
     with the driver, so the driver can make
     /dev/fcrypt work */
  FILE* driver;

  guint64 key;
  gcry_error_t gcry_err;
  gpointer p;
  size_t len;
  gint exit_status;

  gcry_check_version(NULL);

  f_arg_add ("activate", 'a',
	     _("Prepare /dev/fcrypt for use with the"
	       " specified file."), 
	     FALSE,
	     FALSE);

  f_arg_add ("deactivate", 's',
	     _("Umount /dev/fcrypt and stop running this "
	       "program."), 
	     FALSE,
	     FALSE);

  f_arg_add ("init", 'i',
	     _("Initializes a file with random data."), 
	     FALSE,
	     FALSE);

  f_arg_add ("add", 'c',
	     _("Creates a new encrypted partition on file."),
	     FALSE,
	     FALSE);

  f_arg_add ("delete", 'd',
	     _("Delete an encrypted partition and all"
	       " its contents."),
	     FALSE,
	     FALSE);

  f_arg_add ("clean", 'b',
	     _("Subtitue entire contents of a partition"
	       " by random data."),
	     FALSE,
	     FALSE);

  f_arg_add ("manual", 'm',
	     _("Specifies manually place and size"
	       " of the partition."),
	     FALSE,
	     FALSE);

  f_arg_add ("random", 'r',
	     _("Uses a random password to "
	       "encrypt/decrypt. So that "
	       "data is lost after this execution."),
	     FALSE,
	     FALSE);

  f_arg_add ("paranoid", 'p',
	     _("Take paranoid security"
	     " measures. Do not forget"
	       " that this program is experimental."),
	     FALSE,
	     FALSE);

  f_arg_add ( "better-random", 'n',
	      _("Uses read/write data in order"
		" to improve the \"random\"."),
	      FALSE,
	      FALSE);

#if 0
  f_arg_add ( "hidden-max", 'j',
	      _("Mixture (xor) the unused space"
		" on the partition with random data."),
	      FALSE,
	      FALSE );
#endif

#if HAVE_GTHREAD
  f_arg_add( "auto-entropy", 'y',
	     _("Mixture (xor) the seed and the output"
	       " with /dev/random"),
	     FALSE,
	     FALSE );
#endif

  lrandom_file = fopen( "/dev/random", "r");

  /* I do not think 0 is a good seed, I do not think
     it's a good idea to use at any time. We say that
     the seed is exhausted to generate a new one (which
     is really randomly). */
  random_data.used = random_dur;
  
  additional_information( TRUE,
			  "<gonzalfj@ymail.com>",
			  "http://www.gnu.org/software/fcrypt" );

  f_arg_process( argc, argv,
		 "FCrypt 0.1.01",
		 "2012 Fabio J. Gonzalez "
		 "<gonzalfj@ymail.com>",
		 "A simple encryption system for"
		 " GNU/Linux\nto"
		 " be resistant to censorship and "
		 "protect\nyour"
		 " data. "
		 "Remember that file space where"
		 "\nthere are no"
		 " encrypted data will automa-\n"
		 "tically be filled"
		 " with random data.",
		 "<file>" );

  if( f_args_len() < 1 )
    {
      f_arg_help();
      return -1;
    }

#if HAVE_GTHREAD
  if( f_arg_exists("auto-entropy") )
    {
      g_thread_init(NULL);
      g_thread_create( auto_entropy, NULL,
		       FALSE, NULL );
    }
#endif

  if( f_arg_exists("paranoid") )
    random_dur = random_dur/5;

  filename = f_arg_get(0);

  if( sizeof(gsize) != 8 )
    {
      printf("WARNING:Incompatible with 64 bits\n"
	     "pc`s(never your own pc)\n");
    }

  /* 1000 characters should be more
     than sufficient */
  tmp = g_malloc0(1000 * sizeof(char));

  /* The procedure activate, depends
     this file */
  fp = fopen( filename, "r+" );

  if( fp == NULL )
    {
      printf("Can not open file %s\n", filename);
      while(1)
	{
	  show("You want to create it?(yes,no)\n");
	  scanf("%s", tmp);
 
	  if( g_strcmp0(tmp, "no") == 0 )
	    goto error;

	  if( g_strcmp0(tmp, "yes") == 0 )
	    {
	      fp = fopen( filename, "w");
	      fclose(fp);
	      fp = NULL;
	      break;
	    }
	}
    }

  /* must be the first */
  if( fp == NULL || f_arg_exists("init") )
    {
      if(fp == NULL)
	fp = fopen(filename, "r+");

      if(fp == NULL)
	{
	  printf("Can not open file %s\n", filename);
	  goto error;
	}

      printf("Filling header with random data...\n");
      random_message;
      j = CSECTOR_SIZE * HEADER_SIZE;

      for(i = 0; i < j; i++)
	{
	  getrandom( &byte, 1 );

	  if( f_arg_exists("better-random") )
	    rand_fwrite( &byte, 1, 1, fp);
	  else
	    fwrite( &byte, 1, 1, fp );

	  rand_comp(i, j);
	}

      printf("\nFinished random header\n");
    }

  if( !f_arg_exists ("activate") )
    {
      if( f_arg_exists("add") )
	{
	  printf("Write the sector password.\n");
	  printf("(Use letters, numbers and others\n"
		 "characters for best security)\n");

	  c = getpass("password:");

	  if(c)
	    c = g_memdup( c, strlen(c)+1 );

	  gcry_err =
	    gcry_kdf_derive( c, strlen(c), 
			     GCRY_KDF_SIMPLE_S2K,
			     GCRY_MD_MD5, NULL, 0,
			     0, sizeof(key), &key );

	  g_assert(gcry_err == 0);

	  *sec = fgcry_getsector( fp, c,
				  strlen(c) );

	  if( sec->key == key )
	    {
	      printf( _("This key already exists!\n") );
	      return -1;
	    }
	  sec->key = key;

	  if( strlen(c) < 30)
	    printf("WARNING: Less than 30 characters\n");

	  printf("Password to decrypt/encrypt the data.\n"
		 "(leave this field empty to be\n"
		 " the same)\n");

	  c2 = getpass("password:");

	  if(c2)
	    c2 = g_memdup(c2, strlen(c2)+1);

	  if( c2 != NULL && strlen(c2) > 0 )
	    {
	      gcry_err =
		gcry_kdf_derive( c2, strlen(c2),
				 GCRY_KDF_SIMPLE_S2K,
				 GCRY_MD_SHA256, NULL,
				 0, 0, 32,
				 &sec->password[0] );
	      g_assert(gcry_err == 0);
	    }
	  else
	    {
	      gcry_err =
		gcry_kdf_derive( c, strlen(c),
				 GCRY_KDF_SIMPLE_S2K,
				 GCRY_MD_SHA256, NULL,
				 0, 0, 32,
				 &sec->password[0] );
	      g_assert(gcry_err == 0);
	    }

	  sec->pos = 0;
	  sec->len = 0;

	  printf("Write position of new sector:(more"
		 " than %dMB)\n"
		 "Syntax: [integer number][B, KB, MB, "
		 "GB, TB]\n",
		 (HEADER_SIZE * CSECTOR_SIZE) / MEGA(1) );
	  scanf( "%u%s", &sec->pos, tmp);

	  if( g_strcmp0(tmp, "KB") == 0)
	    sec->pos *= KILO(1);
	  else if( g_strcmp0(tmp, "MB") == 0)
	    sec->pos *= MEGA(1);
	  else if( g_strcmp0(tmp, "GB") == 0)
	    sec->pos *= GIGA(1);
	  else if( g_strcmp0(tmp, "TB") == 0)
	    sec->pos *= TERA(1);

	  printf("Write length of new sector:\n");
	  scanf( "%u%s", &sec->len, tmp);

	  if( g_strcmp0(tmp, "KB") == 0)
	    sec->len *= KILO(1);
	  else if( g_strcmp0(tmp, "MB") == 0)
	    sec->len *= MEGA(1);
	  else if( g_strcmp0(tmp, "GB") == 0)
	    sec->len *= GIGA(1);
	  else if( g_strcmp0(tmp, "TB") == 0)
	    sec->len *= TERA(1);

	  if( sec->pos > file_len(fp) )
	    {

	      if(sec->pos > file_len(fp))
		fseek(fp, 0, SEEK_END);
	      else
		fseek( fp, sec->pos, SEEK_SET);

	      if( sec->pos > file_len(fp) )
		{
		  j = sec->pos - file_len(fp);
		  random_message;

		  for( i = 0; i < j; i++)
		    {
		      getrandom( &byte, 1);
		      if(f_arg_exists("better-random") )
			rand_fwrite( &byte, 1, 1, fp);
		      else
			fwrite( &byte, 1, 1, fp);

		      rand_comp(i, j);
		    }
		}

	      for( i = 0; i < sec->len; i++ )
		{
		  getrandom( &byte, 1);
		  if(f_arg_exists("better-random") )
		    rand_fwrite( &byte, 1, 1, fp);
		  else
		    fwrite( &byte, 1, 1, fp);

		  rand_comp(i, sec->len);
		}
	      show("\n");
	    }
	  else
	    {
	      while(1)
		{
		  printf("Replace sector with random"
			 " data?(yes, no)\n");
		  scanf( "%s", tmp );

		  if( g_strcmp0(tmp, "no") == 0)
		    break;

		  if( g_strcmp0(tmp, "yes") == 0)
		    {
		      random_message;
		      
		      fseek( fp, sec->pos, SEEK_SET );
		      for( i = 0; i < sec->len; i++ )
			{
			  getrandom( &byte, sizeof(byte));

			  if(f_arg_exists("better-random") )
			    rand_fwrite( &byte, 1, 1, fp);
			  else
			    fwrite( &byte, 1, 1, fp);

			  rand_comp(i, sec->len);
			}
		      show("\n");
		      break;
		    }
		}
	    }
	  len = fgcry_len(sizeof(Sector));

	  gcry_err =
	    fgcry_encrypt( NULL, c,
			   strlen(c),
			   sec,
			   len );

	  fgcry_err(gcry_err);

	  fread( &k, 1, sizeof(k), lrandom_file);
	  k = k % MEGA(1);

	  fseek( fp, k * len, SEEK_SET);
	  fwrite( sec, 1, len, fp );
	}


      if( f_arg_exists("delete") )
	{
	  sec->pos = 0;
	  sec->len = 0;
	  fseek(fp, 0, SEEK_SET);
	  printf("Write the sector key.\n");

	  c = getpass("password:");

	  *sec = fgcry_getsector( fp, c, strlen(c) );

	  if( sec->key != 0 )
	    {

	      if( fgcry_delsector( fp, c, strlen(c)) )
		g_print(_("Partition information"
			  " successfully deleted.\n"));
	      else
		g_print( _("Partition information not found.\n") );

	      while(1)
		{
		  printf("Delete sector content"
			 " too?(yes, no)");
		  scanf( "%s", tmp);

		  if( g_strcmp0(tmp, "yes") == 0 )
		    {
		      fseek( fp, sec->pos, SEEK_SET);
		      for( j = 0; j < sec->len; j++)
			{
			  getrandom( &byte, sizeof(byte) );

			  if(f_arg_exists("better-random") )
			    rand_fwrite( &byte, 1, 1, fp);
			  else
			    fwrite( &byte, 1, 1, fp);

			  rand_comp(j, sec->len);
			}
		      break;
		    }

		  if( g_strcmp0(tmp, "no") == 0 )
		    break;
		}
	    }
	  else
	    {
	      printf("Sector header not found\n");
	    }
	}

      if(f_arg_exists("clean"))
	{
	  fseek(fp, 0, SEEK_SET);

	  sec->pos = 0;
	  sec->len = 0;

	  gcry_err =
	    gcry_cipher_open( &cipher, GCRY_CIPHER_AES256, 
			      GCRY_CIPHER_MODE_ECB, 0);
	  fgcry_err(gcry_err);

	  show( "Enter the key that identifies"
		" the desired partition.\n");

	  c = getpass("password:");

	  *sec = fgcry_getsector( fp, c, strlen(c) );

	  if( sec->key != 0 )
	    {
	      fseek( fp, sec->pos, SEEK_SET);
	      random_message;
	      for( j = 0; j < sec->len; j++)
		{
		  getrandom( &byte,
			     sizeof(byte));

		  if(f_arg_exists("better-random") )
		    rand_fwrite( &byte, 1, 1, fp);
		  else
		    fwrite( &byte, 1, 1, fp);

		  rand_comp( j, sec->len);
		}
	      g_print("\n");
	    }
	  else
	    {
	      printf("Sector header not found\n");
	    }
	  gcry_cipher_close(cipher);
	}
    } 
  else
    {
      gcry_err =
	gcry_cipher_open( &cipher, GCRY_CIPHER_AES256, 
			  GCRY_CIPHER_MODE_ECB, 0);
      fgcry_err(gcry_err);

      if( g_strcmp0(g_get_user_name(), "root") != 0 )
	printf( _("WARNING: This program not runs as root\n") );

      sec->pos = 0;
      sec->len = 0;
      if( f_arg_exists("manual") )
	{

	  if( !f_arg_exists("random") )
	    c = getpass("password:");

	  printf("Write position of new sector:(more"
		 " than %dMB, if you wish preserve header)\n"
		 "Syntax: [integer number][B, KB, MB, "
		 "GB, TB]\n",
		 (HEADER_SIZE * CSECTOR_SIZE) / MEGA(1) );
	  scanf( "%u%s", &sec->pos, tmp);

	  if( g_strcmp0(tmp, "KB") == 0)
	    sec->pos *= KILO(1);
	  else if( g_strcmp0(tmp, "MB") == 0)
	    sec->pos *= MEGA(1);
	  else if( g_strcmp0(tmp, "GB") == 0)
	    sec->pos *= GIGA(1);
	  else if( g_strcmp0(tmp, "TB") == 0)
	    sec->pos *= TERA(1);

	  printf("Write length of new sector:\n");
	  scanf( "%u%s", &sec->len, tmp);

	  if( g_strcmp0(tmp, "KB") == 0)
	    sec->len *= KILO(1);
	  else if( g_strcmp0(tmp, "MB") == 0)
	    sec->len *= MEGA(1);
	  else if( g_strcmp0(tmp, "GB") == 0)
	    sec->len *= GIGA(1);
	  else if( g_strcmp0(tmp, "TB") == 0)
	    sec->len *= TERA(1);

	  if( !f_arg_exists("random") )
	    {
	      gcry_err =
		gcry_kdf_derive( c, strlen(c),
				 GCRY_KDF_SIMPLE_S2K,
				 GCRY_MD_SHA256, NULL,
				 0, 0, 32,
				 &sec->password[0] );
	    }
	}
      else 
	{
	  c = getpass("password:");

	  *sec = fgcry_getsector( fp, c, strlen(c) );

	  if( sec->key == 0 )
	    {
	      printf( "No partition found "
		      "for this password\n" );
	      return -1;
	    }

	}

      if( f_arg_exists("random") )
	{
	  random_message;
	  if( !f_arg_exists("paranoid") )
	    fread( &sec->password[0], 1, 32, lrandom_file );
	  else
	    {
	      fread( &k, 1, sizeof(k),
		     lrandom_file);

	      k =  4 + (k % 16);
	      for( j = 0; j < k; j++ )
		{
		  for( i = 0; i < 32; i++ )
		    {
		      fread( &byte, 1, sizeof(byte),
			     lrandom_file);
		      sec->password[i] ^= byte;
		    }
		  rand_comp(j, k);
		}
	    }
	  show("\nRandomly password generated!\n");
	}

      if( sec->pos > file_len(fp) )
	{
	  j = sec->pos - file_len(fp);
	  j += sec->len;

	  fseek( fp, file_len(fp), SEEK_SET );

	  random_message;
	  for( i = 0; i < j; i++)
	    {
	      getrandom( &byte, 1 );

	      if(f_arg_exists("better-random") )
		rand_fwrite( &byte, 1, 1, fp);
	      else
		fwrite( &byte, 1, 1, fp);

	      rand_comp( i, j);
	    }
	  show("\n");
	} else if( (sec->pos + sec->len) > file_len(fp) )
	{
	  j = (sec->pos + sec->len) - file_len(fp);
	  fseek( fp, file_len(fp), SEEK_SET );

	  random_message;
	  for( i = 0; i < j; i++)
	    {
	      getrandom( &byte, 1 );

	      if(f_arg_exists("better-random") )
		rand_fwrite( &byte, 1, 1, fp);
	      else
		fwrite( &byte, 1, 1, fp);

	      rand_comp( i, j);
	    }
	  show("\n");
	}

      g_spawn_command_line_sync
	( "sh "
	  KENEL_MODULE
	  "fcrypt_load.sh start",
	  NULL,
	  NULL,
	  &exit_status,
	  NULL );

      if( exit_status != 0 )
	{
	  printf("Unable to initialize the module.\n");
	  return -1;
	}

      driver = fopen("/dev/fcrypt_info", "w+");

      if( driver == NULL )
	{
	  printf("Cannot open /dev/fcrypt_info\n");
	  return -1;
	}

      act_running = (!0);

      signal( SIGKILL, stop );
      signal( SIGTERM, stop );
      signal( SIGABRT, stop );
      signal( SIGQUIT, stop );
      signal( SIGINT, stop );

      while( act_running )
	{
	  /* wait someone use /dev/fcrypt,
	     this generate a request.
	     The len to read, not matter,
	     the driver will send that. */
	  fread( driver_req, 1, sizeof(Request), driver );

	  if( !act_running )
	    {
	      printf("Stoping...\n");
	      break;
	    }

	  if( driver_req->type == REQUEST_SEEK )
	    {
	      if( driver_req->pos < sec->len )
		{
		  fseek( fp, sec->pos + driver_req->pos,
			 SEEK_SET );
		  driver_res->ok = 1;
		}
	      else
		driver_res->ok = 0;

	      driver_res->pos = sec->pos + driver_req->pos;

	      fwrite( driver_res, 1,
		      sizeof(Response),
		      driver );
	    }
	  else if( driver_req->type == REQUEST_READ )
	    {
	      if( driver_req->pos >= sec->len )
		{
		  driver_res->len = 0;
		  fwrite( driver_res, 1,
			  sizeof(Response), driver );
		  continue;
		}

	      if( driver_req->pos+driver_req->len >
		  sec->len )
		driver_res->len = sec->len-driver_req->pos;

	      fseek( fp, sec->pos + driver_req->pos,
		     SEEK_SET );

	      fread( driver_res + sizeof(Response), 1,
		     driver_res->len, fp );
	      fwrite( driver_res, 1, 
		      sizeof(Response) + driver_res->len,
		      driver );
	    }
	  else if( driver_req->type == REQUEST_WRITE )
	    {
	      if( driver_req->pos >= sec->len )
		{
		  driver_res->len = 0;
		  fwrite( driver_res, 1,
			  sizeof(Response), driver );
		  continue;
		}

	      if( driver_req->pos+driver_req->len >
		  sec->len )
		driver_res->len = sec->len-driver_req->pos;

	      fseek( fp, sec->pos + driver_req->pos,
		     SEEK_SET );

	      fwrite( driver_req + sizeof(Request), 1,
		      driver_res->len, fp );

	      fwrite( driver_res, 1,
		      sizeof(Response), driver );
	    }
	}
      
      fclose(driver);

      g_spawn_command_line_sync( "sh "
				 KENEL_MODULE
				 "fcrypt_load.sh stop",
				 NULL,
				 NULL,
				 &exit_status,
				 NULL);
      gcry_cipher_close(cipher);
    }

  if( f_arg_exists("deactivate") )
    {

      g_spawn_command_line_sync
	("umount /dev/fcrypt",
	 NULL, NULL, NULL,
	 NULL);

      g_spawn_command_line_sync
	("killall fcrypt_user",
	 NULL, NULL, NULL,
	 NULL);

      g_spawn_command_line_sync
	("killall fcrypt_setup",
	 NULL, NULL, NULL,
	 NULL);
    }

 success:
  g_free(tmp);
  return 0;
 
 error:
  g_free(tmp);
  return -1;
}
