/*
  FCrypt - A censorship-resistant cryptosystem
  Copyright (C) 2012  Fabio J. Gonzalez <gonzalfj@ymail.com>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/genhd.h>
#include <linux/slab.h>
#include <linux/semaphore.h>
#include "../common.h"

MODULE_LICENSE("GPL");

struct DriverRequest
{
  enum
    {
      DREQUEST_READ,
      DREQUEST_WRITE,
      DREQUEST_SEEK
    } type;

  char __user *data;
  size_t len;
  size_t pos;
  loff_t offset;
};

static struct DriverRequest dr;

static dev_t dcrypt;
static dev_t dinfo;

static struct  file_operations bfcrypt;
static struct file_operations binfo;

static struct cdev* cdev_crypt;
static struct cdev* cdev_info;

static struct semaphore have_request;
static struct semaphore working;
static struct semaphore response_needed;
static struct semaphore waiting_response;
static Request user_request;
static Response user_response;

void* buf;

#define fdown(x, value)				\
  if( down_interruptible(x) != 0 )		\
    return value;

ssize_t
info_write( struct file *file, char __user *user,
	    size_t len, loff_t *offset)
{
  fdown(&response_needed, 0);

  if( dr.type == DREQUEST_SEEK )
    {
      copy_from_user( &user_response, user,
		      sizeof(user_response));
      up(&waiting_response);
      return sizeof(user_response);
    }
  else if( dr.type == DREQUEST_READ )
    {
      copy_from_user( &user_response, user,
		      sizeof(user_response));

      copy_from_user( buf,
		      user + sizeof(user_response),
		      user_response.len);

      up(&waiting_response);
      return sizeof(user_response) + user_response.len;
    }
  else if( dr.type == DREQUEST_WRITE )
    {
      copy_from_user( &user_response, user,
		      sizeof(user_response));

      up(&waiting_response);
      return user_response.len;
    }
  
}

ssize_t
info_read( struct file *file, char __user *user,
	   size_t len, loff_t *offset)
{  
  fdown(&have_request, 0);

  if( dr.type == DREQUEST_SEEK )
    {
      user_request.type = REQUEST_SEEK;
      user_request.pos = dr.offset;
      copy_to_user( user, &user_request,
		    sizeof(user_request) );
      up(&response_needed);
      return sizeof(user_request);
    } 
  else if( dr.type == DREQUEST_READ )
    {
      user_request.type = REQUEST_READ;
      user_request.pos = dr.pos;
      if(dr.len > 1024)
	user_request.len = 1024;
      else
	user_request.len = dr.len;
      copy_to_user( user, &user_request,
		    sizeof(user_request) );
      up(&response_needed);
      return sizeof(user_request);
    }
  else if( dr.type == DREQUEST_WRITE )
    {
      user_request.type = REQUEST_WRITE;
      user_request.pos = dr.pos;
      user_request.len = dr.len;

      copy_to_user( user, &user_request,
		    sizeof(user_request) );
      copy_to_user( user+sizeof(user_request),
		    buf, user_request.len );
      up(&response_needed);
      return sizeof(user_request) + user_request.len;
    }
}

loff_t
crypt_seek (struct file *file, loff_t offset, int n)
{
  loff_t ret;

  fdown(&working, -1);

  dr.type = DREQUEST_SEEK;
  dr.offset = offset;
  up(&have_request);
  down(&waiting_response);

  if( user_response.ok != 0 )
    ret = user_response.pos;
  else
    ret = -1;
  up(&working);

  return ret;
}

ssize_t
crypt_read( struct file *file, char __user *user,
	    size_t len, loff_t *offset)
{
  ssize_t tmp1;

  fdown(&working, 0);

  dr.type = DREQUEST_READ;
  dr.data = user;
  dr.len = len;
  dr.pos = *offset;
  up(&have_request);
  down(&waiting_response);
  copy_to_user( user, buf,
		user_response.len );
  *offset += user_response.len;
  tmp1 = user_response.len;
  up(&working);
  return tmp1;
}

ssize_t
crypt_write( struct file *file, char __user *user,
	    size_t len, loff_t *offset)
{
  ssize_t ret;

  fdown(&working, 0);

  dr.data = user;

  if( dr.len > 1024 )
    dr.len = 1024;
  else
    dr.len = len;

  dr.pos = *offset;
  dr.type = DREQUEST_WRITE;

  copy_from_user( buf, user,
		  dr.len );

  up(&have_request);
  down(&waiting_response);

  ret = user_response.len;
  *offset += ret;
  up(&working);
  return ret;
}


static int
f_init(void)
{
  sema_init( &have_request, 0);
  sema_init( &working, 1);
  sema_init( &response_needed, 0);
  sema_init( &waiting_response, 0);

  buf = kmalloc(1024, GFP_KERNEL);

  alloc_chrdev_region( &dcrypt, 0, 1, "fcrypt");
  alloc_chrdev_region( &dinfo, 0, 1, "fcrypt_info");

  bfcrypt.owner = THIS_MODULE;
  bfcrypt.llseek = crypt_seek;
  bfcrypt.read = crypt_read;
  bfcrypt.write = crypt_write;

  binfo.owner = THIS_MODULE;
  binfo.read = info_read;
  binfo.write = info_write;

  cdev_crypt = cdev_alloc();
  cdev_crypt->ops = &bfcrypt;
  cdev_crypt->owner = THIS_MODULE;

  cdev_info = cdev_alloc();
  cdev_info->ops = &binfo;
  cdev_info->owner = THIS_MODULE;

  cdev_add(cdev_crypt, dcrypt, 1);
  cdev_add(cdev_info, dinfo, 1);
}

static void
f_exit(void)
{

  kfree(buf);
  cdev_del(cdev_crypt);
  cdev_del(cdev_info);

  unregister_chrdev_region( dinfo, 1);
  unregister_chrdev_region( dcrypt, 1);
}

module_init(f_init);
module_exit(f_exit);
